/*
    UPN-Calc - An arbitrary precision RPN calculator for J2ME/MIDP1
    Copyright (C) 2005 Kai G. Schwebke, upn.schwebke.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
 * RPN Calculator Widget - display, input/output and program logic
 */


import javax.microedition.lcdui.*;


class UPNCanvas
    extends Canvas
    implements CommandListener
{
   private final UPNMIDlet midlet;
   private final Display display;
   private final boolean useColor;
   private final Command exitCommand;
   private final Command delCommand;
   private final Command optionsCommand;
   private final Command helpCommand;
   private final Command aboutCommand;

   private final List optionsMenu;
   private final Command optionsMenuBack;

   private final List angleMeasureList;
   private final Command angleMeasureBack;

   private final List fontSizeList;
   private final Command fontSizeBack;

   private final Form precForm;
   private final TextField precItemStack;
   private final TextField precItemInternal;
   private final Command precFormOK;

   private final Alert helpAlert;
   private final Alert aboutAlert;

   private Stack stack; 
   private String message;
   private String angleMeasure;
   private String fontSize;

   private final String tooFewArgMsg;
   private final String wrongArgMsg;

   private String line = new String();

   // 0: Stack Display
   // 1: Line Editing
   // 2: Command Select
   // 3: Func Select
   private int mode;

   public UPNCanvas(UPNMIDlet midlet, Display display)
   {
      stack = new Stack();
      mode = 0;
      message = "welcome to UPN";
      angleMeasure = "RAD"; // DEG/RAD/GRAD
      fontSize = "SMALL";

      tooFewArgMsg = "too few args";
      wrongArgMsg = "wrong arg";

      this.midlet = midlet;
      this.display = display;

      useColor = display.isColor();

      setCommandListener(this);

      exitCommand = new Command("Exit", Command.EXIT, 1);
      delCommand = new Command("DEL", Command.ITEM, 1);
      optionsCommand = new Command("Options", Command.ITEM, 2);
      helpCommand = new Command("Help", Command.HELP, 2);
      aboutCommand = new Command("About", Command.ITEM, 2);
      addCommand(exitCommand);
      addMainMenu();

      angleMeasureList = new List("Angle Measure", List.IMPLICIT);
      angleMeasureList.append("DEG", null);
      angleMeasureList.append("RAD", null);
      angleMeasureList.append("GRAD", null);
      angleMeasureBack = new Command("Cancel", Command.CANCEL, 1);
      angleMeasureList.addCommand(angleMeasureBack);
      angleMeasureList.setCommandListener(this);

      fontSizeList = new List("Font Size", List.IMPLICIT);
      fontSizeList.append("SMALL", null);
      fontSizeList.append("MEDIUM", null);
      fontSizeList.append("LARGE", null);
      fontSizeBack = new Command("Cancel", Command.CANCEL, 1);
      fontSizeList.addCommand(fontSizeBack);
      fontSizeList.setCommandListener(this);

      optionsMenu = new List("Options", List.IMPLICIT);
      optionsMenu.append("Angle Measure", null);
      optionsMenu.append("Precision", null);
      optionsMenu.append("Font Size", null);
      optionsMenuBack = new Command("Cancel", Command.CANCEL, 1);
      optionsMenu.addCommand(optionsMenuBack);
      optionsMenu.setCommandListener(this);

      precForm = new Form("Precision");
      precItemStack = new TextField("Stack", "", 3, TextField.NUMERIC);
      precItemInternal = new TextField("Internal", "", 3, TextField.NUMERIC);
      precForm.append(precItemStack);
      precForm.append(precItemInternal);
      precFormOK = new Command("Back", Command.BACK, 1);
      precForm.addCommand(precFormOK);
      precForm.setCommandListener(this);

      helpAlert = new Alert("UPN Help",
         "Reverse Polnish Notation Calculator\n\n"+
	 "Enter values using numeric keys; # for "+
	 "-, ., E.\n"+
	 "Use * to access operations and functions.\n"+
	 "For further help see manual at upn.schwebke.com.",
	 null, AlertType.INFO);
      helpAlert.setTimeout(Alert.FOREVER);

      aboutAlert = new Alert("UPN Help",
         "UPN 1.0.0\n\n"+
	 "Copyright (C) 2005 Kai G. Schwebke\n"+
	 "UPN comes with ABSOLUTELY NO WARRANTY; "+
         "This is free software, and you are welcome "+
	 "to redistribute it under certain conditions; "+
	 "see COPYING for details.",
	 null, AlertType.INFO);
      aboutAlert.setTimeout(Alert.FOREVER);
   }

   public String getAngleMeasure()
   {
      return angleMeasure;
   }

   public void setAngleMeasure(String am)
   {
      angleMeasure = am;
      repaint();
   }

   public String getFontSize()
   {
      return fontSize;
   }

   public void setFontSize(String fs)
   {
      fontSize = fs;
      repaint();
   }

   public int getStackPrec()
   {
      return stack.prec;
   }

   public void setStackPrec(int p)
   {
      stack.prec = p;
   }

   public int getDefaultPrec()
   {
      return BCDFloat.getDefaultPrec();
   }

   public void setDefaultPrec(int p)
   {
      BCDFloat.setDefaultPrec(p);
   }

   public Stack getStack()
   {
      return stack;
   }

   protected void addMainMenu()
   {
      addCommand(optionsCommand);
      addCommand(helpCommand);
      addCommand(aboutCommand);
   }

   protected void removeMainMenu()
   {
      removeCommand(optionsCommand);
      removeCommand(helpCommand);
      removeCommand(aboutCommand);
   }

    public void paint(Graphics g)
    {
	if (fontSize.equals("MEDIUM"))
	{
	   g.setFont(Font.getFont(0, 0, Font.SIZE_MEDIUM));
	} else if (fontSize.equals("LARGE")) {
	   g.setFont(Font.getFont(0, 0, Font.SIZE_LARGE));
	} else {
	   g.setFont(Font.getFont(0, 0, Font.SIZE_SMALL));
	}

        int width = getWidth();
        int height = getHeight();
	Font f = g.getFont();
	int lineHeight = f.getHeight();
	int y = height-2-lineHeight;
	int s = stack.size()-1;

        g.setColor(255, 255, 255);
        g.fillRect(0, 0, width, height);
        g.setColor(0, 0, 0);
	g.drawRect(0, 0, width - 1, height - 1);
	g.fillRect(0, 0, width, lineHeight+3);
        g.setClip(1, 1, width - 2, height - 2);

	if (mode == 1)
	{
	   g.drawString(line, width-2, y, g.TOP|g.RIGHT);
	   y -= lineHeight;
	}

	while ( (s >= 0) && (y >= 4+lineHeight) )
	{
	   g.drawString(stack.at(s).toString(), width-2, y, g.TOP|g.RIGHT);
	   g.drawString(Integer.toString(stack.size()-s)+":",
			   2, y, g.TOP|g.LEFT);
	   --s;
	   y -= lineHeight;
	}

	g.setColor(255, 255, 255);
	if (message.length() == 0)
	{
	   g.drawString("SS: "+stack.size()+"   "+angleMeasure,
			  2, 2, g.TOP|g.LEFT);
	} else {
	   g.drawString(message, 2, 2, g.TOP|g.LEFT);
	}

	// show Command Selector
	if ((mode == 2) || (mode == 3))
	{
	   int border = 5;
	   int pad = 2;
	   g.setColor(0, 0, 0);
	   g.drawRect(border+1, border+1, width-2*border, height-2*border);
	   g.setColor(255, 255, 255);
	   g.fillRect(border, border, width-2*border, height-2*border);
	   g.setColor(0, 0, 0);
	   g.drawRect(border, border, width-2*border, height-2*border);
	   g.setClip(border+1, border+1, width-border-2, height-border-2);

	   int gridX = (width-2*(border+pad))/3+2;
	   int gridY = (height-2*(border+pad))/4;

	   int x0 = border+pad;
	   int y0 = border+pad+1;

	   if (mode == 2)
	   {
	      g.drawString("1 +", x0+0*gridX, y0+0*gridY, g.TOP|g.LEFT);
	      g.drawString("2 -", x0+1*gridX, y0+0*gridY, g.TOP|g.LEFT);
	      g.drawString("3 DUP", x0+2*gridX, y0+0*gridY, g.TOP|g.LEFT);

	      g.drawString("4 *", x0+0*gridX, y0+1*gridY, g.TOP|g.LEFT);
	      g.drawString("5 /", x0+1*gridX, y0+1*gridY, g.TOP|g.LEFT);
	      g.drawString("6 SWP", x0+2*gridX, y0+1*gridY, g.TOP|g.LEFT);

	      g.drawString("7 x", x0+0*gridX, y0+2*gridY, g.TOP|g.LEFT);
	      g.drawString("8 sqr", x0+1*gridX, y0+2*gridY, g.TOP|g.LEFT);
	      g.drawString("9 DRP", x0+2*gridX, y0+2*gridY, g.TOP|g.LEFT);

	      g.drawString("* ENT", x0+0*gridX, y0+3*gridY, g.TOP|g.LEFT);
	      g.drawString("0 f(x)", x0+1*gridX, y0+3*gridY, g.TOP|g.LEFT);
	      g.drawString("# CLR", x0+2*gridX, y0+3*gridY, g.TOP|g.LEFT);
	   }
	   if (mode == 3)
	   {
	      g.drawString("1 inv", x0+0*gridX, y0+0*gridY, g.TOP|g.LEFT);
	      g.drawString("2 exp", x0+1*gridX, y0+0*gridY, g.TOP|g.LEFT);
	      g.drawString("3 ln", x0+2*gridX, y0+0*gridY, g.TOP|g.LEFT);

	      g.drawString("4 sin", x0+0*gridX, y0+1*gridY, g.TOP|g.LEFT);
	      g.drawString("5 cos", x0+1*gridX, y0+1*gridY, g.TOP|g.LEFT);
	      g.drawString("6 tan", x0+2*gridX, y0+1*gridY, g.TOP|g.LEFT);

	      g.drawString("7 asin", x0+0*gridX, y0+2*gridY, g.TOP|g.LEFT);
	      g.drawString("8 acos", x0+1*gridX, y0+2*gridY, g.TOP|g.LEFT);
	      g.drawString("9 atan", x0+2*gridX, y0+2*gridY, g.TOP|g.LEFT);

	      g.drawString("* ENT", x0+0*gridX, y0+3*gridY, g.TOP|g.LEFT);
	      g.drawString("0 Pi", x0+1*gridX, y0+3*gridY, g.TOP|g.LEFT);
	      g.drawString("# pow", x0+2*gridX, y0+3*gridY, g.TOP|g.LEFT);
	   }
	}
    }

    protected void keyPressed(int keyCode)
    {
       message = "";
       int prevMode = mode;
       handleKey(keyCode);
       if ((prevMode != 1) && (mode == 1))
       {
	  addCommand(delCommand);
	  removeMainMenu();
       }
       if ((prevMode == 1) && (mode != 1))
       {
	  removeCommand(delCommand);
	  addMainMenu();
       }
       repaint();
    }

    private void handleKey(int keyCode)
    {
       if ((mode == 0) || (mode == 1))
       {
	  if (keyCode == KEY_NUM0)
	  {
	     mode = 1;
	     line = line + '0';
	     return;
	  }
	  if (keyCode == KEY_NUM1)
	  {
	     mode = 1;
	     line = line + '1';
	     return;
	  }
	  if (keyCode == KEY_NUM2)
	  {
	     mode = 1;
	     line = line + '2';
	     return;
	  }
	  if (keyCode == KEY_NUM3)
	  {
	     mode = 1;
	     line = line + '3';
	     return;
	  }
	  if (keyCode == KEY_NUM4)
	  {
	     mode = 1;
	     line = line + '4';
	     return;
	  }
	  if (keyCode == KEY_NUM5)
	  {
	     mode = 1;
	     line = line + '5';
	     return;
	  }
	  if (keyCode == KEY_NUM6)
	  {
	     mode = 1;
	     line = line + '6';
	     return;
	  }
	  if (keyCode == KEY_NUM7)
	  {
	     mode = 1;
	     line = line + '7';
	     return;
	  }
	  if (keyCode == KEY_NUM8)
	  {
	     mode = 1;
	     line = line + '8';
	     return;
	  }
	  if (keyCode == KEY_NUM9)
	  {
	     mode = 1;
	     line = line + '9';
	     return;
	  }
	  if (keyCode == KEY_POUND)
	  {
	     mode = 1;
	     if (line.length() == 0)
	     {
		line = "-";
	     } else if (line.indexOf("E") != (-1)) {
		line = line + "-";
	     } else if (line.indexOf(".") != (-1)) {
		line = line + "E";
	     } else {
		line = line + '.';
	     }
	     return;
	  }

       }

       if (mode == 1)
       {
	  if (keyCode == KEY_STAR)
	  {
	     mode = 2;
	     try {
		if (line.length() > 0)
		{
		   stack.push(new BCDFloat(line));
		}
	     } catch (NumberFormatException e) {
		stack.pop();
		mode = 1;
		message = "parse err";
	     }
	     if (mode == 2)
	     {
		line = "";
	     }
	     return;
	  }
       }

       if (mode == 0)
       {
	  if (keyCode == KEY_STAR)
	  {
	     mode = 2;
	     return;
	  }
       }

       if (mode == 2) // Command ...
       {
	  mode = 0;
	  displayWaitMsg();
	  if (keyCode == KEY_NUM1) // +
	  {
	     if (stack.size() >= 2)
	     {
		BCDFloat b = stack.top();
		stack.pop();
		BCDFloat a = stack.top();
		stack.pop();
		stack.push(a.add(b));
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM2) // -
	  {
	     if (stack.size() >= 2)
	     {
		BCDFloat b = stack.top();
		stack.pop();
		BCDFloat a = stack.top();
		stack.pop();
		stack.push(a.sub(b));
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM3) // DUP
	  {
	     if (stack.size() >= 1)
	     {
		stack.push(stack.top().copy());
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM4) // *
	  {
	     if (stack.size() >= 2)
	     {
		BCDFloat b = stack.top();
		stack.pop();
		BCDFloat a = stack.top();
		stack.pop();
		stack.push(a.mul(b));
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM5) // /
	  {
	     if (stack.size() >= 2)
	     {
		BCDFloat b = stack.top();
		stack.pop();
		BCDFloat a = stack.top();
		stack.pop();
		stack.push(a.div(b));
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM6) // SWAP
	  {
	     if (stack.size() >= 2)
	     {
		BCDFloat b = stack.top();
		stack.pop();
		BCDFloat a = stack.top();
		stack.pop();
		
		stack.push(b);
		stack.push(a);
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM7) // x
	  {
	     if (stack.size() >= 1)
	     {
		BCDFloat a = stack.top();
		stack.pop();
		stack.push(a.mul(a));
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM8) // SQRT
	  {
	     if (stack.size() >= 1)
	     {
		BCDFloat a = stack.top();
		stack.pop();
		stack.push(a.sqrt());
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM9) // DROP
	  {
	     if (stack.size() >= 1)
	     {
		stack.pop();
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_STAR) // ENT
	  {
	  }
	  if (keyCode == KEY_NUM0) // f(x)...
	  {
	     mode = 3;
	  }
	  if (keyCode == KEY_POUND) // CLR
	  {
	     stack.clear();
	  }
	  return;
       }

       if (mode == 3) // Command ...
       {
	  mode = 0;
	  displayWaitMsg();
	  if (keyCode == KEY_NUM1) // 1/x
	  {
	     if (stack.size() >= 1)
	     {
		BCDFloat a = stack.top();
		stack.pop();
		stack.push((new BCDFloat("1")).div(a));
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM2) // e^x
	  {
	     if (stack.size() >= 1)
	     {
		BCDFloat a = stack.top();
		stack.pop();
		stack.push(a.exp());
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM3) // ln x
	  {
	     if (stack.size() >= 1)
	     {
		BCDFloat a = stack.top();
		stack.pop();
		stack.push(a.ln());
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM4) // sin x
	  {
	     if (stack.size() >= 1)
	     {
		BCDFloat a = angleToRad(stack.top());
		stack.pop();
		stack.push(a.sin());
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM5) // cos x
	  {
	     if (stack.size() >= 1)
	     {
		BCDFloat a = angleToRad(stack.top());
		stack.pop();
		stack.push(a.cos());
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM6) // tan x
	  {
	     if (stack.size() >= 1)
	     {
		BCDFloat a = angleToRad(stack.top());
		stack.pop();
		stack.push(a.tan());
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM7) // asin
	  {
	     if (stack.size() >= 1)
	     {
		BCDFloat a = stack.top();
		stack.pop();
		stack.push(radToAngle(a.asin()));
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM8)
	  {
	     if (stack.size() >= 1) // acos
	     {
		BCDFloat a = stack.top();
		stack.pop();
		stack.push(radToAngle(a.acos()));
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_NUM9) // atan
	  {
	     if (stack.size() >= 1)
	     {
		BCDFloat a = stack.top();
		stack.pop();
		stack.push(radToAngle(a.atan()));
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  if (keyCode == KEY_STAR)
	  {
	  }
	  if (keyCode == KEY_NUM0)
	  {
	     stack.push(BCDConst.getPi());
	  }
	  if (keyCode == KEY_POUND) // y^x
	  {
	     if (stack.size() >= 2)
	     {
		BCDFloat b = stack.top();
		stack.pop();
		BCDFloat a = stack.top();
		stack.pop();
		stack.push(a.ln().mul(b).exp());
		stack.top().normalize(a.e*b.toInteger());
	     } else {
		message = tooFewArgMsg;
	     }
	  }
	  return;
       }
    }

    private void displayWaitMsg()
    {
       message = "please wait...";
       repaint();
       serviceRepaints();
       message = "";
    }


    public void commandAction(Command c, Displayable d)
    {
        if (c == exitCommand)
        {
            midlet.exitRequested();
        }
	if (c == delCommand)
	{
	   if (line.length() > 0)
	   {
	      line = line.substring(0, line.length()-1);
	      repaint();
	   }
	}
	if (c == optionsCommand)
	{
	   display.setCurrent(optionsMenu);
	}
	if ((c == optionsMenu.SELECT_COMMAND) && (d == optionsMenu))
	{
	   int i = optionsMenu.getSelectedIndex();

	   if (i == 0)
	   {
	      commandAngleMeasure();
	   }

	   if (i == 1)
	   {
	      commandPrec();
	   }

	   if (i == 2)
	   {
	      commandFontSize();
	   }
	}
	if (c == optionsMenuBack)
	{
	   display.setCurrent(this);
	}
	if ((c == angleMeasureList.SELECT_COMMAND) && (d == angleMeasureList))
	{
	   int i = angleMeasureList.getSelectedIndex();
	   if (i == 0)
	   {
	      angleMeasure = "DEG";
	   }
	   if (i == 1)
	   {
	      angleMeasure = "RAD";
	   }
	   if (i == 2)
	   {
	      angleMeasure = "GRAD";
	   }

	   display.setCurrent(this);
	}
	if (c == angleMeasureBack)
	{
	   display.setCurrent(this);
	}
	if ((c == fontSizeList.SELECT_COMMAND) && (d == fontSizeList))
	{
	   int i = fontSizeList.getSelectedIndex();
	   if (i == 0)
	   {
	      fontSize = "SMALL";
	   }
	   if (i == 1)
	   {
	      fontSize = "MEDIUM";
	   }
	   if (i == 2)
	   {
	      fontSize = "LARGE";
	   }

	   display.setCurrent(this);
	}
	if (c == fontSizeBack)
	{
	   display.setCurrent(this);
	}
	if (c == precFormOK)
	{
	   try {
	      stack.prec =
		 Integer.parseInt(precItemStack.getString());
	      BCDFloat.setDefaultPrec(
		 Integer.parseInt(precItemInternal.getString()));
	   } catch (NumberFormatException e) {
	   }
	   display.setCurrent(this);
	}
	if (c == helpCommand)
	{
	   display.setCurrent(helpAlert, this);
	}
	if (c == aboutCommand)
	{
	   display.setCurrent(aboutAlert, this);
	}
    }

    public void commandAngleMeasure()
    {
	if (angleMeasure.equals("DEG"))
	{
	   angleMeasureList.setSelectedIndex(0, true);
	}
	if (angleMeasure.equals("RAD"))
	{
	   angleMeasureList.setSelectedIndex(1, true);
	}
	if (angleMeasure.equals("GRAD"))
	{
	   angleMeasureList.setSelectedIndex(2, true);
	}

	display.setCurrent(angleMeasureList);
    }

    public void commandFontSize()
    {
	if (fontSize.equals("SMALL"))
	{
	   fontSizeList.setSelectedIndex(0, true);
	}
	if (fontSize.equals("MEDIUM"))
	{
	   fontSizeList.setSelectedIndex(1, true);
	}
	if (fontSize.equals("LARGE"))
	{
	   fontSizeList.setSelectedIndex(2, true);
	}

	display.setCurrent(fontSizeList);
    }

    public void commandPrec()
    {
	precItemStack.setString(Integer.toString(stack.prec));
	precItemInternal.setString(Integer.toString(BCDFloat.getDefaultPrec()));
	display.setCurrent(precForm);
    }

    public void start()
    {
    }

    public void stop()
    {
    }

    public BCDFloat angleToRad(BCDFloat a)
    {
       if (angleMeasure.equals("RAD"))
       {
	  return a;
       }
       if (angleMeasure.equals("DEG"))
       {
	  a = a.div(new BCDFloat("360"));
       }
       if (angleMeasure.equals("GRAD"))
       {
	  a = a.div(new BCDFloat("400"));
       }
       return a.mul(BCDConst.get2Pi());
    }

    public BCDFloat radToAngle(BCDFloat a)
    {
       if (angleMeasure.equals("RAD"))
       {
	  return a;
       }
       if (angleMeasure.equals("DEG"))
       {
	  a = a.mul(new BCDFloat("360"));
       }
       if (angleMeasure.equals("GRAD"))
       {
	  a = a.mul(new BCDFloat("400"));
       }
       return a.div(BCDConst.get2Pi());
    }

}
